# -*- coding: utf-8 -*-


import asyncio
import inspect
import os
import signal as _signal
import sys
import traceback

from asyncio import (
    AbstractEventLoop,
    StreamReader,
    StreamWriter,
)
from asyncio import Queue  # noqa: F401
from contextlib import contextmanager
from itertools import count
from unittest import mock
from typing import (
    Callable,
    Iterable,
    Iterator,
    Optional,
    TextIO,
    Tuple,
    Union,
)

from ._mempipe import mempipe


def call_with_minimal_args(f: Callable, **kwds):
    spec = inspect.signature(f)
    kwds = {
        k: kwds[k] for k, p in spec.parameters.items()
        if p.kind != inspect.Parameter.VAR_KEYWORD
    }
    return f(**kwds)


class OutputStreamAdapter(StreamWriter):
    """StreamWriter that wraps a file-like object."""

    def __init__(self, stream: TextIO) -> None:
        self._stream = stream

    def write(self, data: bytes) -> None:
        self._stream.write(data.decode('utf-8'))

    def writelines(self, lines: Iterable[bytes]) -> None:
        for line in lines:
            self._stream.write(line.decode('utf-8'))
            self._stream.write('\n')

    def can_write_eof(self) -> bool:
        return True

    # TODO: test this without closing the "real" sys.stdout!
    def write_eof(self) -> None:
        # self._stream.close()
        pass

    # NOTE: impossible to please `mypy` because documented signature is
    #       incomplete.
    async def drain(self):
        self._stream.flush()


class Process(object):
    """Mock for ``asyncio.subprocess.Process``."""

    def __init__(self, *,
                 pid: int,
                 run: Callable,
                 argv=[],
                 env=None,
                 kwds={},
                 stdin: Optional[int]=None,
                 stdout: Union[int, TextIO, None]=None,
                 stderr: Union[int, TextIO, None]=None,
                 limit: Optional[int]=None,
                 loop: Optional[AbstractEventLoop]=None) -> None:

        self._loop = loop or asyncio.get_event_loop()
        self._pid = pid
        self._stdin = None  # type: Optional[StreamWriter]
        self._stdout = None  # type: Optional[StreamReader]
        self._stderr = None  # type: Optional[StreamReader]

        # Handle standard input redirection.
        r_stdin = None  # Optional[StreamReader]
        if stdin == asyncio.subprocess.PIPE:
            r_stdin, self._stdin = mempipe(limit=limit, loop=loop)
        else:
            # TODO: wrap `sys.stdin` in a `StreamReader`.
            r_stdin, self._stdin = None, None

        # Handle standard output redirection.
        if stdout == asyncio.subprocess.PIPE:
            self._stdout, w_stdout = mempipe(limit=limit, loop=loop)
        else:
            stdout = stdout or sys.stdout
            assert stdout is not None
            assert not isinstance(stdout, int)
            self._stdout, w_stdout = None, OutputStreamAdapter(stdout)

        # Handle standard error redirection.
        if stderr == asyncio.subprocess.PIPE:
            self._stderr, w_stderr = mempipe(limit=limit, loop=loop)
        else:
            stderr = stderr or sys.stderr
            assert stderr is not None
            assert not isinstance(stderr, int)
            self._stderr, w_stderr = None, OutputStreamAdapter(stderr)

        # Mock signal handling.
        self._signals = asyncio.Queue()  # type: Queue

        # Start the application-defined process simulation.
        self._done = asyncio.Event(loop=loop)
        self._task = self._loop.create_task(self._run_wrapper(
            run,
            stdin=r_stdin,
            stdout=w_stdout,
            stderr=w_stderr,
            signals=self._signals,
            env=env or {k: v for k, v in os.environ.items()},
            argv=argv,
            kwds=kwds,
        ))

        # Keep a reference to the streams, we'll need them later.
        self._w_stdout = w_stdout
        self._w_stderr = w_stderr

        # Process exit code is undefined until the simulation completes.
        self._returncode = None  # type: Optional[int]

    async def _run_wrapper(self, run: Callable,
                           *, stdout, stderr, **kwds) -> int:
        try:
            return await call_with_minimal_args(
                run, stdout=stdout, stderr=stderr, **kwds
            )
        except asyncio.CancelledError:
            return 1
        finally:
            await stdout.drain()
            await stderr.drain()
            self._done.set()

    @property
    def pid(self) -> int:
        return self._pid

    @property
    def stdin(self) -> Optional[StreamWriter]:
        return self._stdin

    @property
    def stdout(self) -> Optional[StreamReader]:
        return self._stdout

    @property
    def stderr(self) -> Optional[StreamReader]:
        return self._stderr

    async def wait(self) -> int:
        await asyncio.wait({self._task})
        await self._done.wait()
        e = self._task.exception()
        if e is None:
            r = self._task.result()
            if r is None:
                r = 0
            self._returncode = r
        else:
            # Format traceback and send it to stderr (as if it had been printed
            # in the child process' output).
            self._w_stderr.writelines(
                line.encode('utf-8')
                for line in traceback.format_exception(
                    e.__class__, e, e.__traceback__
                )
            )
            self._returncode = 1
        assert self._w_stdout
        assert self._w_stderr
        self._w_stdout.write_eof()
        self._w_stderr.write_eof()
        return self._returncode

    async def communicate(self, input: bytes=b'') -> Tuple[Optional[bytes],
                                                           Optional[bytes]]:
        if self._stdin:
            self._stdin.write(input)
            self._stdin.write_eof()
        await self.wait()
        stdout = None
        if self._stdout:
            stdout = await self._stdout.read()
        stderr = None
        if self._stderr:
            stderr = await self._stderr.read()
        return stdout, stderr

    def send_signal(self, signal: int) -> None:
        self._signals.put_nowait(signal)

    def terminate(self) -> None:
        self._task.cancel()

    def kill(self) -> None:
        if sys.platform == 'win32':
            self.terminate()
        else:
            # NOTE: for a real process, we'd send SIGKILL, which would then be
            #       passed as SIGINT to the application, but we don't have a
            #       kernel to make that substution here.
            self.send_signal(_signal.SIGINT)

    @property
    def returncode(self) -> Optional[int]:
        return self._returncode


@contextmanager
def mock_subprocess(run: Callable,
                    loop: Optional[AbstractEventLoop]=None) -> Iterator[None]:
    """Calls ``run()`` instead of spawning a sub-process.

    :param run: A coroutine function that simulates the sub-process.  Can
     return ``None`` or ``0`` to simulate successful process execution or a
     non-zero error code to simulate sub-process terminate with a non-zero exit
     code.  If an exception is raised, the result is 1 (non-zero).  This
     function can accept a variable number of arguments, see below.

    Dependency injection is used with the ``run()`` coroutine function to pass
    only arguments that are declared in the function's signature.  Omit all but
    the arguments you intend to use.  Here are all the available arguments:

    - ``argv``: a list of strings passed as positional arguments to
      ``asyncio.create_subprocess_exec()``.
    - ``stdin``: an ``asyncio.StreamReader`` instance.  When output is not
      redirected, this reads from the "real" ``sys.stdin``.
    - ``stdout``: an ``asyncio.StreamWriter`` instance.  When output is not
      redirected, this writes to the "real" ``sys.stdout``.
    - ``stderr``: an ``asyncio.StreamWriter`` instance.  When output is not
      redirected, this writes to the "real" ``sys.stderr``.
    - ``env``: a ``dict`` containing environment variables passed to
      ``asyncio.create_subprocess_exec()``.
    - ``signals``: an ``asyncio.Queue`` object that receives integers passed to
      ``asyncio.Process.send_signal()``.
    - ``kwds``: extra keyword arguments passed to
      ``asyncio.create_subprocess_exec()``.

    .. versionadded:: 0.1

    """

    loop = loop or asyncio.get_event_loop()

    pid = count(start=1)

    def create_subprocess_exec(*args, stdin=None, stdout=None, env=None,
                               stderr=None, loop=None, limit=None, **kwds):
        """Mock for ``asyncio.create_subprocess_exec()``."""
        loop = loop or asyncio.get_event_loop()
        f = asyncio.Future()
        process = Process(
            pid=next(pid),
            run=run,
            loop=loop,
            argv=list(args),
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            env=env,
            limit=limit,
            kwds=kwds,
        )
        loop.call_soon(f.set_result, process)
        return f

    with mock.patch('asyncio.create_subprocess_exec', create_subprocess_exec):
        yield None



# -*- coding: utf-8 -*-


import asyncio

from asyncio import Future  # noqa: F401
from asyncio import StreamReader
from asyncio import StreamWriter
from asyncio import Task  # noqa: F401
from typing import Any  # noqa: F401
from typing import Callable
from typing import Optional
from typing import Set  # noqa: F401

from ._cancel import wait_until_cancelled
from ._typing import SocketServer


def _effective_bind_port(server: SocketServer) -> int:
    return server.sockets[0].getsockname()[1]


async def tcp_server(**kwds)-> None:
    """Run a TCP server in the foreground.

    **Note**: this function is a coroutine.

    This context manager provides the following features over direct use of
    ``asyncio.create_server()``:

    - runs in the foreground, blocking the calling task
    - automatically closes the server object with graceful shutdown.

    There are two main difference with the related :py:class:`~TCPServer`:

    - based on ``asyncio.create_server()`` for compatibility with libraries
      that are based on protocols (asyncio's callback API)
    - runs in the foreground

    .. versionadded:: 0.3

    """

    loop = kwds.pop('loop', None) or asyncio.get_event_loop()

    server = await loop.create_server(**kwds)
    try:
        await wait_until_cancelled(loop=loop)
    finally:
        server.close()
        await server.wait_closed()


class TCPServer(object):
    """Asynchronous context manager to accept TCP connections.

    This context manager provides the following features over direct use of
    ``asyncio.start_server()``:

    - connection handlers are coroutines
    - prevent leaking connections when the writer is not properly closed by the
      connection handler
    - automatically cancel all tasks that handle connections when it's time to
      shut down
    - wait until all connections are closed before shutting down the server
      application (includes handling of a rare race condition)

    :param host: Network interface on which to listen for incoming connections.

    :param port: Port number on which to listen for incoming connections.  You
     can pass a value of zero to ask the system to pick an available port for
     you.  After ``wait_started()``, the ``port`` property will contain the
     port number on which the server is listening.
    :param callback: Coroutine function that will be used to spawn a task for
     each established connection.  This coroutine must accept two positional
     arguments: ``(reader, writer)`` which allow interaction with the peer.
    :param loop: Event loop in which to run the server's asynchronous tasks.
     When ``None``, the current default event loop will be used.

    .. versionadded:: 0.2

    """

    def __init__(self, host: str, port: int,
                 callback: Callable, loop=None) -> None:
        self._callback = callback
        self._host = host
        self._port = port
        self._loop = loop or asyncio.get_event_loop()
        self._sessions = set()  # type: Set[Task]
        self._boot = None  # type: Optional[Task]
        self._server = None  # type: Optional[SocketServer]

    @property
    def host(self) -> str:
        """Network interface on which the server listens for connections.

        See:

        - :py:data:`aiotk.TCPServer.port`
        """
        return self._host

    @property
    def port(self) -> int:
        """Port number on which the server listens for connections.

        See:

        - :py:data:`aiotk.TCPServer.host`

        .. versionchanged:: 0.4.1

           When asking the system to choose the available port, this property
           now returns the actual port number (rather than zero).

        """
        return self._port

    # NOTE: cannot declare type for return value here because the class
    #       definition is not completed...
    async def __aenter__(self):
        self.start()
        await self.wait_started()
        return self

    async def __aexit__(self, *args) -> Optional[bool]:
        self.close()
        await self.wait_closed()
        # NOTE: `mypy` will not allow `return` without an expression even when
        #       it accepts `None`...
        return None

    def start(self) -> None:
        """Start accepting connections.

        Only use this method if you are not using the server as an asynchronous
        context manager.

        See:

        - :py:meth:`aiotk.TCPServer.wait_started`
        - :py:meth:`aiotk.TCPServer.close`
        """
        if self._boot:
            raise Exception('Already running.')
        self._boot = self._loop.create_task(asyncio.start_server(
            self._client_connected, host=self._host, port=self._port,
        ))

    # NOTE: start is synchronous, so we can't await in there.
    async def wait_started(self) -> None:
        """Wait until the server is ready to accept connections.

        See:

        - :py:meth:`aiotk.TCPServer.start`
        - :py:meth:`aiotk.TCPServer.close`
        """
        if not self._boot:
            raise Exception('Not started.')
        if self._server is None:
            self._server = await self._boot
            assert self._server is not None
            self._port = _effective_bind_port(self._server)

    # NOTE: cancel pending sessions immediately.
    #
    # TODO: make cancelling session optionnal (graceful shutdown).
    def close(self) -> None:
        """Stop accepting connections.

        .. note::

           Since connections may still be pending in the kernel's TCP stack at
           the time where you call this, it's possible that new connections
           seem to get established after you signal your intent to stop
           accepting connections.

        See:

        - :py:meth:`aiotk.TCPServer.wait_started`
        - :py:meth:`aiotk.TCPServer.wait_closed`

        """
        if self._boot and not self._boot.done():
            raise Exception('Starting.')
        if not self._server:
            return
        self._server.close()
        for session in self._sessions:
            session.cancel()

    # NOTES:
    # - caller can retry if they hit the timeout;
    # - not clear if we can still accept connections until
    #   ```self._server.wait_closed()` completes, so we cancel again after it
    #   does.
    async def wait_closed(self, timeout=None) -> None:
        """Wait until all connections are closed.

        See:

        - :py:meth:`aiotk.TCPServer.wait_started`
        - :py:meth:`aiotk.TCPServer.wait_closed`
        """

        if not self._server:
            return
        await self._server.wait_closed()
        for session in self._sessions:
            session.cancel()
        pending = {t for t in self._sessions}  # type: Set[Future[Any]]
        if pending:
            _, pending = await asyncio.wait(
                pending,
                return_when=asyncio.ALL_COMPLETED,
                timeout=timeout,
            )
        if pending:
            raise asyncio.TimeoutError()
        assert not self._sessions

    # NOTE: not closing writer here leaks connection.
    async def _client_wrapper(self,
                              reader: StreamReader,
                              writer: StreamWriter) -> None:
        try:
            return await self._callback(
                reader=reader, writer=writer,
            )
        except asyncio.CancelledError:
            pass
        finally:
            writer.close()

    # NOTE: exception here leaks connection.
    def _client_connected(self,
                          reader: StreamReader,
                          writer: StreamWriter):
        s = self._loop.create_task(
            self._client_wrapper(reader, writer)
        )
        s.add_done_callback(self._sessions.remove)
        self._sessions.add(s)



# -*- coding: utf-8 -*-


import asyncio
import os

from asyncio import AbstractEventLoop  # noqa: F401
from asyncio import Future  # noqa: F401
from asyncio import StreamReader
from asyncio import StreamWriter
from asyncio import Task  # noqa: F401
from typing import Any  # noqa: F401
from typing import Callable
from typing import Optional
from typing import Set  # noqa: F401
from typing import Union  # noqa: F401

from ._typing import Server  # noqa: F401


class UnixSocketServer(object):
    """Asynchronous context manager to accept UNIX connections.

    This context manager provides the following features over direct use of
    ``asyncio.start_unix_server()``:

    - connection handlers are coroutines
    - prevent leaking connections when the writer is not properly closed by the
      connection handler
    - automatically cancel all tasks that handle connections when it's time to
      shut down
    - wait until all connections are closed before shutting down the server
      application (includes handling of a rare race condition)
    - automatically unlink the UNIX socket when shutting down (assuming the
      process does not crash)

    :param path: Path to the UNIX socket on which to listen for incoming
     connections.
    :param callback: Coroutine function that will be used to spawn a task for
     each established connection.  This coroutine must accept two positional
     arguments: ``(reader, writer)`` which allow interaction with the peer.
    :param loop: Event loop in which to run the server's asynchronous tasks.
     When ``None``, the current default event loop will be used.

    .. versionadded:: 0.1

    """

    def __init__(self, path: str, callback: Callable,
                 loop: Optional[AbstractEventLoop]=None) -> None:
        self._callback = callback
        self._path = path
        self._loop = loop or asyncio.get_event_loop()
        self._sessions = set()  # type: Set[Task]
        self._boot = None  # type: Optional[Task]
        self._server = None  # type: Optional[Server]

    @property
    def path(self) -> str:
        """UNIX socket on which the server listens for incoming connections."""
        return self._path

    # NOTE: cannot declare type for return value here because the class
    #       definition is not completed...
    async def __aenter__(self):
        self.start()
        await self.wait_started()
        return self

    async def __aexit__(self, *args) -> None:
        self.close()
        await self.wait_closed()

    def start(self) -> None:
        """Start accepting connections.

        Only use this method if you are not using the server as an asynchronous
        context manager.

        See:

        - :py:meth:`aiotk.TCPServer.wait_started`
        - :py:meth:`aiotk.TCPServer.close`
        """
        if self._boot:
            raise Exception('Already running.')
        self._boot = self._loop.create_task(asyncio.start_unix_server(
            self._client_connected, path=self._path, loop=self._loop,
        ))

    # NOTE: start is synchronous, so we can't await in there.
    async def wait_started(self) -> None:
        """Wait until the server is ready to accept connections.

        See:

        - :py:meth:`aiotk.TCPServer.start`
        - :py:meth:`aiotk.TCPServer.close`
        """
        if not self._boot:
            raise Exception('Not started.')
        if self._server is None:
            self._server = await self._boot

    # NOTE: cancel pending sessions immediately.
    def close(self) -> None:
        """Stop accepting connections.

        .. note::

           Since connections may still be pending in the kernel's TCP stack at
           the time where you call this, it's possible that new connections
           seem to get established after you signal your intent to stop
           accepting connections.

        See:

        - :py:meth:`aiotk.TCPServer.wait_started`
        - :py:meth:`aiotk.TCPServer.wait_closed`

        """
        # NOTE: there is a natural race here and we could not manage to
        #       reliably reach this branch while the boot task had not yet
        #       completed.
        if self._boot and not self._boot.done():
            raise Exception('Starting.')  # pragma: no cover
        if not self._server:
            return
        self._server.close()
        for session in self._sessions:
            session.cancel()

    # NOTES:
    # - caller can retry if they hit the timeout;
    # - not clear if we can still accept connections until
    #   ```self._server.wait_closed()` completes, so we cancel again after it
    #   does.
    async def wait_closed(self, timeout: Optional[float]=None) -> None:
        """Wait until all connections are closed.

        See:

        - :py:meth:`aiotk.TCPServer.wait_started`
        - :py:meth:`aiotk.TCPServer.wait_closed`
        """
        # NOTE: there is a natural race here and we could not manage to
        #       reliably reach this branch while the boot task had not yet
        #       completed.
        if self._boot and not self._boot.done():
            raise Exception('Starting.')  # pragma: no cover
        if not self._server:
            return
        await self._server.wait_closed()
        for session in self._sessions:
            session.cancel()
        # NOTE: `mypy` does not recognize `Set[Task[Any]]` as a subtype of
        #       `Set[Future[Any]]`.  The only way to resolve this is to create
        #       a copy...
        pending = {t for t in self._sessions}  # type: Set[Future[Any]]
        if pending:
            _, pending = await asyncio.wait(
                pending,
                return_when=asyncio.ALL_COMPLETED,
                timeout=timeout,
            )
        if pending:
            raise asyncio.TimeoutError()
        assert not self._sessions
        if os.path.exists(self._path):
            os.unlink(self._path)

    # NOTE: not closing writer here leaks connection.
    async def _client_wrapper(self,
                              reader: StreamReader,
                              writer: StreamWriter) -> None:
        try:
            return await self._callback(
                reader=reader, writer=writer,
            )
        except asyncio.CancelledError:
            pass
        finally:
            writer.close()

    # NOTE: exception here leaks connection.
    def _client_connected(self,
                          reader: StreamReader,
                          writer: StreamWriter) -> None:
        s = self._loop.create_task(
            self._client_wrapper(reader, writer)
        )
        s.add_done_callback(self._sessions.remove)
        self._sessions.add(s)



# -*- coding: utf-8 -*-


import asyncio

from asyncio import AbstractEventLoop
from contextlib import contextmanager
from typing import Callable, Iterator, Optional


@contextmanager
def reader(fd, callback: Callable,
           loop: Optional[AbstractEventLoop]=None) -> Iterator[None]:
    """Register a low-level reader for a file descriptor.

    **Note**: the proactor event loop does not support readers.

    It's an easy mistake to forget a call to ``.remove_reader()`` and end up
    receiving extra I/O you were not interested in.  This context manager makes
    sure you never forget.

    :param fd: File descriptor to watch for read events.
    :param callback: Called when the file descriptor is ready to read from.
     See asyncio's documentation on ``loop.add_reader()`` for details.
    :param loop: Loop in which the watch will be registered.  Defaults to the
     current event loop.

    This context manager yields nothing.

    .. versionadded:: 0.4

    """

    loop = loop or asyncio.get_event_loop()

    loop.add_reader(fd, callback)
    try:
        yield
    finally:
        loop.remove_reader(fd)



# -*- coding: utf-8 -*-


import asyncio

from asyncio import AbstractEventLoop
from typing import Awaitable

from ._cancel import (
    cancel,
    cancel_all,
    follow_through,
    wait_until_cancelled,
)
from ._ctrlc import handle_ctrlc
from ._io import reader
from ._mempipe import mempipe
from ._monkey import monkey_patch
from ._pool import PoolClosed, TaskPool
from ._posix import UnixSocketServer
from ._sched import PeriodicTask
from ._stack import AsyncExitStack, EnsureDone
from ._tcp import TCPServer, tcp_server
from ._testing import mock_subprocess
from ._udp import udp_server, udp_socket


def run_until_complete(coro: Awaitable, loop: AbstractEventLoop=None):
    """Run a task through to completion.

    The ``.run_until_complete()`` method on asyncio event loop objects does not
    finish tasks when it receives a SIGINT/CTRL-C.  The method simply raises a
    ``KeyboardInterrupt`` exception and this usually results in warnings about
    unfinished tasks plus some "event loop closed" ``RuntimeError`` exceptions
    in pending tasks.

    This is a really annoying default behavior and this function aims at
    replacing that behavior with something that ensures the task actually runs
    through to completion.  When the ``KeyboardInterrupt`` exception is caught,
    the task is canceled and resumed to give it a chance to clean up properly.

    .. versionadded:: 0.4
    .. versionchanged:: 0.5 Can now be called with a ``asyncio.Task`` argument.

    """

    loop = loop or asyncio.get_event_loop()
    if isinstance(coro, asyncio.Task):
        task = coro
    else:
        task = loop.create_task(coro)
    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt:
        task.cancel()
        try:
            loop.run_until_complete(task)
        except asyncio.CancelledError:
            return None
    return task.result()


__all__ = [
    'AsyncExitStack',
    'cancel',
    'cancel_all',
    'EnsureDone',
    'follow_through',
    'handle_ctrlc',
    'mempipe',
    'mock_subprocess',
    'monkey_patch',
    'PeriodicTask',
    'PoolClosed',
    'reader',
    'run_until_complete',
    'TaskPool',
    'TCPServer',
    'tcp_server',
    'udp_server',
    'udp_socket',
    'UnixSocketServer',
    'wait_until_cancelled',
]



# -*- coding: utf-8 -*-


import asyncio
import logging

from aiotk import cancel
from asyncio import AbstractEventLoop
from asyncio import Task  # noqa: F401
from typing import Callable, Optional


class PeriodicTask:
    """Periodically run a coroutine in the background.

    .. code-block:: python

       async def my_task():
           pass

       async with PeriodicTask(my_task, 30.0):
           # do something for a long time.

    .. versionadded:: 0.5

    """

    def __init__(self, func: Callable,
                 interval: float,
                 loop: Optional[AbstractEventLoop]=None) -> None:
        self._loop = loop or asyncio.get_event_loop()
        self._func = func
        self._ival = interval
        self._task = None  # type: Optional[Task]

    # NOTE: cannot declare type for return value here because the class
    #       definition is not completed...
    async def __aenter__(self):
        """Schedule the background task."""
        assert self._task is None
        self._task = self._loop.create_task(self._run())
        return self

    async def __aexit__(self, *args) -> None:
        """Stop the background task and wait for it to complete."""
        assert self._task
        await cancel(self._task)
        self._task = None

    async def _run(self) -> None:
        """Periodically run the task and wait until the schedule."""
        while True:
            try:
                await self._func()
            except asyncio.CancelledError:
                raise
            except Exception:
                logging.exception('Executing periodic task.')
            await asyncio.sleep(self._ival)



# -*- coding: utf-8 -*-


from typing import Any, List
from typing_extensions import Protocol


class Server(Protocol):
    """Asyncio event loop does not publish an interface for server objects."""

    def close(self) -> None:
        ...  # pragma: no cover

    async def wait_closed(self) -> None:
        ...  # pragma: no cover


class SocketServer(Server):
    """Asyncio event loop does not publish an interface for server objects."""

    @property
    def sockets(self) -> List[Any]:
        ...  # pragma: no cover



# -*- coding: utf-8 -*-


import asyncio

from asyncio import AbstractEventLoop, Task
from typing import Any, Optional, Set


async def wait_until_cancelled(*,
                               propagate: bool=True,
                               loop: Optional[AbstractEventLoop]=None) -> None:
    """Wait until the calling task is canceled.

    **Note**: this function is a coroutine.

    When using a context manager to complete one or more background tasks, it's
    common to have the "main" task block until something cancels it (e.g. a
    SIGINT/CTRL-C handler).

    It's also convenient in tests that verify the behavior of cancellation to
    need to spawn a background task that waits forever.

    This leads to the not-so-idiomatic:

    .. code-block:: python

       await asyncio.Future()

    This of often wrapped in a helper function to make the call more readable.
    Instead of propagating multiple variants of this, it should be placed in a
    library that everybody can import.

    :param loop: Loop in which the coroutine will block.  Defaults to the
     current event loop.

    .. versionadded:: 0.3

    """

    loop = loop or asyncio.get_event_loop()

    if propagate:
        await asyncio.Future(loop=loop)
    else:
        try:
            await asyncio.Future(loop=loop)
        except asyncio.CancelledError:
            pass


async def cancel(task: Task, loop: Optional[AbstractEventLoop]=None) -> None:
    """Cancel a task and wait until it's done.

    **Note**: this function is a coroutine.

    Canceling a child task and returning without waiting for the child task to
    complete is a common cause of "event loop closed" ``RuntimeError``
    exceptions, especially during program shutdown.  Therefore, this becomes a
    common pattern:

    .. code-block:: python

        task.cancel()
        await asyncio.wait({task})

    However, if the parent task itself is also canceled, then the
    ``asyncio.wait()`` call will be interrupted and the child task will still
    not complete.  To solve this, we must also manage to trap the
    ``asyncio.CancelledError`` exception and call ``asyncio.wait({task})``
    again and properly re-raise the ``asyncio.CancelledError`` exception.  For
    example:


    .. code-block:: python

        task.cancel()
        try:
            await asyncio.wait({task})
        except asyncio.CancelledError:
            await asyncio.wait({task})
            raise

    This is not trivial and must be done so many times in a program that
    cancels tasks that it merits a replacement API for ``task.cancel()``.

    :param task: The ``asyncio.Task`` object to cancel.
    :param loop: The event loop to use for awaiting.  Defaults to the current
     event loop.

    .. versionadded:: 0.3

    """

    loop = loop or asyncio.get_event_loop()

    task.cancel()
    try:
        await asyncio.wait({task}, loop=loop)
    except asyncio.CancelledError:
        await asyncio.wait({task}, loop=loop)
        raise


async def cancel_all(tasks: Set[Task],
                     loop: Optional[AbstractEventLoop]=None) -> None:
    """Cancel a set of tasks and wait until they're done.

    **Note**: this function is a coroutine.

    Canceling a set of child tasks and returning without waiting for the child
    task to complete is a common cause of "event loop closed" ``RuntimeError``
    exceptions, especially during shutdown of servers with one ore more task
    per connection.  Therefore, this becomes a common pattern:

    .. code-block:: python

        for task in tasks:
            task.cancel()
        await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)

    However, if the parent task itself is also canceled, then the
    ``asyncio.wait()`` call will be interrupted and one or more of the child
    tasks will still not complete.  To solve this, we must also manage to trap
    the ``asyncio.CancelledError`` exception and call ``asyncio.wait(tasks)``
    again and properly re-raise the ``asyncio.CancelledError`` exception.  For
    example:

    .. code-block:: python

        for task in tasks:
            task.cancel()
        try:
            await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
        except asyncio.CancelledError:
            await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
            raise

    This is not trivial and must be done so many server programs that cancels
    tasks that it merits a helper.

    :param tasks: The set of ``asyncio.Task`` objects to cancel.
    :param loop: The event loop to use for awaiting.  Defaults to the current
     event loop.

    .. versionadded:: 0.3

    """

    loop = loop or asyncio.get_event_loop()

    for task in tasks:
        task.cancel()
    try:
        await asyncio.wait(tasks, loop=loop)
    except asyncio.CancelledError:
        await asyncio.wait(tasks, loop=loop)
        raise


async def follow_through(task: Task,
                         loop: Optional[AbstractEventLoop]=None) -> Any:
    """Wait for a task to complete (even if canceled while waiting).

    **Note**: this function is a coroutine.

    Not propagating cancellation to a child task and returning without waiting
    for the child task to complete is a common cause of "event loop closed"
    ``RuntimeError`` exceptions, especially during program shutdown.
    Therefore, this becomes a common pattern:

    .. code-block:: python

        try:
            await asyncio.wait({task})
        except asyncio.CancelledError:
            task.cancel()
            await asyncio.wait({task})
            raise
        return task.result()

    This is not trivial and must be done so many times in a program that spawns
    child tasks that it merits a helper method.

    :param task: The ``asyncio.Task`` object to see through to completion.
    :param loop: The event loop to use for awaiting.  Defaults to the current
     event loop.

    .. versionadded:: 0.3

    """

    loop = loop or asyncio.get_event_loop()

    try:
        await asyncio.wait({task}, loop=loop)
    except asyncio.CancelledError:
        await cancel(task, loop=loop)
        raise
    return task.result()



# -*- coding: utf-8 -*-


from setuptools import find_packages, setup


def readfile(path):
    with open(path, 'rb') as stream:
        return stream.read().decode('utf-8')


readme = readfile('README.rst')


setup(
    name='aiotk',
    maintainer='Andre Caron',
    maintainer_email='andre.l.caron@gmail.com',
    version='0.5.0',
    url='https://github.com/AndreLouisCaron/aiotk',
    packages=find_packages(where='src'),
    package_dir={
        '': 'src',
    },
    package_data={
        'aiotk': [
        ],
    },
    description='aiotk: asyncio toolkit',
    long_description=readme,
    install_requires = [
        'typing-extensions>=3.6.5',
    ],
)



# -*- coding: utf-8 -*-


import asyncio

from asyncio import AbstractEventLoop, StreamReader, StreamWriter
from typing import Iterable, Optional, Tuple


_DEFAULT_LIMIT = 2 ** 16


class _MemoryTransport(asyncio.Transport):
    """Direct connection between a StreamWriter and StreamReader."""

    def __init__(self, reader: StreamReader) -> None:
        super().__init__()
        self._reader = reader

    def write(self, data: bytes) -> None:
        self._reader.feed_data(data)

    def writelines(self, data: Iterable[bytes]) -> None:
        for line in data:
            self._reader.feed_data(line)
            self._reader.feed_data(b'\n')

    def write_eof(self) -> None:
        self._reader.feed_eof()

    def can_write_eof(self) -> bool:
        return True

    def is_closing(self) -> bool:
        return False

    def close(self) -> None:
        self.write_eof()


def mempipe(loop: Optional[AbstractEventLoop]=None,
            limit: int=None) -> Tuple[StreamReader, StreamWriter]:
    """In-memory pipe, returns a ``(reader, writer)`` pair.

    .. versionadded:: 0.1

    """

    loop = loop or asyncio.get_event_loop()
    limit = limit or _DEFAULT_LIMIT

    reader = asyncio.StreamReader(loop=loop, limit=limit)  # type: StreamReader
    writer = asyncio.StreamWriter(
        transport=_MemoryTransport(reader),
        protocol=asyncio.StreamReaderProtocol(reader, loop=loop),
        reader=reader,
        loop=loop,
    )  # type: StreamWriter
    return reader, writer
